@@ -24,19 +24,14 @@ let debug = Trace.debug "copy"
 let protect f g =
   try
     f ()
-  with Sys_error _ | Unix.Unix_error _ | Util.Transient _ as e ->
+  with e ->
     begin try g () with Sys_error _  | Unix.Unix_error _ -> () end;
     raise e
 
 let lwt_protect f g =
   Lwt.catch f
     (fun e ->
-       begin match e with
-         Sys_error _ | Unix.Unix_error _ | Util.Transient _ ->
-           begin try g () with Sys_error _  | Unix.Unix_error _ -> () end
-       | _ ->
-           ()
-       end;
+       begin try g () with Sys_error _  | Unix.Unix_error _ -> () end;
        Lwt.fail e)
 
 (****)
@@ -47,13 +42,14 @@ let lwt_protect f g =
 let checkForChangesToSourceLocal
       fspathFrom pathFrom archDesc archFp archStamp archRess newFpOpt paranoid =
   (* Retrieve attributes of current source file *)
-  let sourceInfo = Fileinfo.get true fspathFrom pathFrom in
+  let sourceInfo = Fileinfo.getBasicWithRess true fspathFrom pathFrom in
+  let sourceType = sourceInfo.Fileinfo.typ in
   match newFpOpt with
     None ->
       (* no newfp provided: so we need to compare the archive with the
          current source *)
       let clearlyChanged =
-           sourceInfo.Fileinfo.typ <> `FILE
+           sourceType <> `FILE
         || Props.length sourceInfo.Fileinfo.desc <> Props.length archDesc
         || Osx.ressLength sourceInfo.Fileinfo.osX.Osx.ressInfo <>
            Osx.ressLength archRess    in
@@ -72,7 +68,7 @@ let checkForChangesToSourceLocal
                              None dataClearlyUnchanged   in
       if dataClearlyUnchanged && ressClearlyUnchanged then begin
         if paranoid && not (Os.isPseudoFingerprint archFp) then begin
-          let newFp = Os.fingerprint fspathFrom pathFrom sourceInfo in
+          let newFp = Os.fingerprint fspathFrom pathFrom sourceType in
           if archFp <> newFp then begin
             Update.markPossiblyUpdated fspathFrom pathFrom;
             raise (Util.Transient (Printf.sprintf
@@ -85,7 +81,7 @@ let checkForChangesToSourceLocal
         end
       end else if
            clearlyChanged
-        || archFp <> Os.fingerprint fspathFrom pathFrom sourceInfo
+        || archFp <> Os.fingerprint fspathFrom pathFrom sourceType
       then
         raise (Util.Transient (Printf.sprintf
           "The source file %s\nhas been modified during synchronization.  \
@@ -96,7 +92,7 @@ let checkForChangesToSourceLocal
       assert (Os.isPseudoFingerprint archFp);
       (* ... so we can't compare the archive with the source; instead we
          need to compare the current source to the new fingerprint: *)
-      if newfp <> Os.fingerprint fspathFrom pathFrom sourceInfo then
+      if newfp <> Os.fingerprint fspathFrom pathFrom sourceType then
         raise (Util.Transient (Printf.sprintf
           "Current source file %s\n not same as transferred file.  \
            Transfer aborted."
@@ -146,8 +142,8 @@ let checkForChangesToSource
 (****)
 
 let fileIsTransferred fspathTo pathTo desc fp ress =
-  let info = Fileinfo.get false fspathTo pathTo in
-  (info,
+  let info = Fileinfo.getBasicWithRess false fspathTo pathTo in
+  (Fileinfo.basic info,
    info.Fileinfo.typ = `FILE
      &&
    Props.length info.Fileinfo.desc = Props.length desc
@@ -155,7 +151,7 @@ let fileIsTransferred fspathTo pathTo desc fp ress =
    Osx.ressLength info.Fileinfo.osX.Osx.ressInfo =
    Osx.ressLength ress
      &&
-   let fp' = Os.fingerprint fspathTo pathTo info in
+   let fp' = Os.fingerprint fspathTo pathTo info.Fileinfo.typ in
    fp' = fp)
 
 (* We slice the files in 1GB chunks because that's the limit for
@@ -209,13 +205,13 @@ type transferStatus251 =
   | TransferFailed of string
 
 type transferStatus =
-    TransferSucceeded of Fileinfo.t
-  | TransferNeedsDoubleCheckAgainstCurrentSource of Fileinfo.t * Os.fullfingerprint
+    TransferSucceeded of Fileinfo.basic
+  | TransferNeedsDoubleCheckAgainstCurrentSource of Fileinfo.basic * Os.fullfingerprint
   | TransferFailed of string
 
 let mtransferStatus = Umarshal.(sum3
-                                  Fileinfo.m
-                                  (prod2 Fileinfo.m Os.mfullfingerprint id id)
+                                  Fileinfo.mbasic
+                                  (prod2 Fileinfo.mbasic Os.mfullfingerprint id id)
                                   string
                                   (function
                                    | TransferSucceeded a -> I31 a
@@ -249,8 +245,8 @@ let transferStatus_of_compat251 (st : transferStatus251) : transferStatus =
    calculated the current source fingerprint.
  *)
 let paranoidCheck fspathTo pathTo realPathTo desc fp ress =
-  let info = Fileinfo.get false fspathTo pathTo in
-  let fp' = Os.fingerprint fspathTo pathTo info in
+  let info = Fileinfo.getBasic false fspathTo pathTo in
+  let fp' = Os.fingerprint fspathTo pathTo info.Fileinfo.typ in
   if Os.isPseudoFingerprint fp then begin
     Lwt.return (TransferNeedsDoubleCheckAgainstCurrentSource (info,fp'))
   end else if fp' <> fp then begin
@@ -284,11 +280,9 @@ let saveTempFileLocal (fspathTo, (pathTo, realPathTo, reason)) =
         reason
         (Fspath.toDebugString (Fspath.concat fspathTo savepath))))
 
-let convV0 = Remote.makeConvV0FunRet Fileinfo.to_compat251 Fileinfo.of_compat251
-
 let saveTempFileOnRoot =
-  Remote.registerRootCmd "saveTempFile" ~convV0
-    Umarshal.(prod3 Path.mlocal Path.mlocal string id id) Fileinfo.m
+  Remote.registerRootCmd "saveTempFile"
+    Umarshal.(prod3 Path.mlocal Path.mlocal string id id) Umarshal.unit
     saveTempFileLocal
 
 (****)
@@ -300,7 +294,18 @@ let removeOldTempFile fspathTo pathTo =
     Os.delete fspathTo pathTo
   end
 
-let openFileIn fspath path kind =
+(* There is an issue that not all threads are immediately cancelled when there
+   is a connection error. A waiting thread (in this case probably a thread in
+   one of the Lwt regions) may have been started and could open an fd but may
+   never be able to complete. [protect], [lwt_protect] and any other cleanup
+   code may never be triggered in this scenario because the thread just stops
+   (as eventually the connection cleanup kicks in and all threads are stopped).
+   As a hacky(?) solution, keep track of all open fds and close them when the
+   connection breaks. *)
+let inFdResource = Remote.resourceWithConnCleanup close_in close_in_noerr
+let outFdResource = Remote.resourceWithConnCleanup close_out close_out_noerr
+
+let openFileIn' fspath path kind =
   match kind with
     `DATA ->
       Fs.open_in_bin (Fspath.concat fspath path)
@@ -311,12 +316,19 @@ let openFileIn fspath path kind =
   | `RESS ->
       Osx.openRessIn fspath path
 
-let openFileOut fspath path kind len =
+let openFileIn fspath path kind =
+  inFdResource.register (openFileIn' fspath path kind)
+
+let closeFileIn = inFdResource.release
+
+let closeFileInNoErr = inFdResource.release_noerr
+
+let openFileOut' fspath path kind len =
   match kind with
     `DATA ->
       let fullpath = Fspath.concat fspath path in
-      let flags = [Unix.O_WRONLY;Unix.O_CREAT] in
-      let perm = 0o600 in
+      let flags = [Unix.O_WRONLY; Unix.O_CREAT; Unix.O_CLOEXEC] in
+      let perm = if Prefs.read Props.dontChmod then Props.perms Props.fileDefault else 0o600 in
       begin match Util.osType with
         `Win32 ->
           Fs.open_out_gen
@@ -336,7 +348,7 @@ let openFileOut fspath path kind len =
       end
   | `DATA_APPEND len ->
       let fullpath = Fspath.concat fspath path in
-      let perm = 0o600 in
+      let perm = if Prefs.read Props.dontChmod then Props.perms Props.fileDefault else 0o600 in
       let ch = Fs.open_out_gen [Open_wronly; Open_binary] perm fullpath in
       if not (Prefs.read Props.dontChmod) then Fs.chmod fullpath perm;
       LargeFile.seek_out ch (Uutil.Filesize.toInt64 len);
@@ -344,6 +356,13 @@ let openFileOut fspath path kind len =
   | `RESS ->
       Osx.openRessOut fspath path len
 
+let openFileOut fspath path kind len =
+  outFdResource.register (openFileOut' fspath path kind len)
+
+let closeFileOut = outFdResource.release
+
+let closeFileOutNoErr = outFdResource.release_noerr
+
 let setFileinfo fspathTo pathTo realPathTo update desc =
   match update with
     `Update _ -> Fileinfo.set fspathTo pathTo (`Copy realPathTo) desc
@@ -351,6 +370,57 @@ let setFileinfo fspathTo pathTo realPathTo update desc =
 
 (****)
 
+(* This unfortunate complexity is here to reduce network round-trips
+   and calls to [Update.translatePath], primarily in [Files.setProp]. *)
+let mxpath = Umarshal.(sum2 Path.mlocal Path.m)
+               (function `Local p -> I21 p | `Global p -> I22 p)
+               (function I21 p -> `Local p | I22 p -> `Global p)
+
+let loadPropsExtDataLocal (fspath, path, desc) =
+  let localPath = match path with
+    | `Local p -> p
+    | `Global p -> Update.translatePathLocal fspath p in
+  (Some localPath, Props.loadExtData fspath localPath desc)
+
+let loadPropsExtDataOnServer = Remote.registerServerCmd "propsExtData"
+  Umarshal.(prod3 Fspath.m mxpath Props.m id id)
+  Umarshal.(prod2 (option Path.mlocal) Props.mx id id)
+  (fun connFrom args -> Lwt.return (loadPropsExtDataLocal args))
+
+let propsWithExtDataLocal fspath path desc =
+  try (None, Props.withExtData desc)
+  with Not_found -> loadPropsExtDataLocal (fspath, path, desc)
+
+let propsWithExtDataConn connFrom fspath path desc =
+  try Lwt.return (None, Props.withExtData desc)
+  with Not_found -> loadPropsExtDataOnServer connFrom (fspath, path, desc)
+
+let propsExtDataOnRoot root path desc =
+  match root with
+  | (Common.Local, fspath) ->
+      Lwt.return (propsWithExtDataLocal fspath path desc)
+  | (Remote _, fspath) ->
+      propsWithExtDataConn (Remote.connectionOfRoot root) fspath path desc
+
+let propsWithExtData connFrom fspath path desc =
+  propsWithExtDataConn connFrom fspath (`Local path) desc >>= fun x ->
+  Lwt.return (snd x)
+
+let readPropsExtData root path desc =
+  propsExtDataOnRoot root (`Local path) desc >>= fun x ->
+  Lwt.return (snd x)
+
+let readPropsExtDataG root path desc =
+  propsExtDataOnRoot root (`Global path) desc
+
+(****)
+
+(* The fds opened in this function normally shouldn't be tracked for extra
+   cleanup at connection close because this is sequential non-Lwt code. Yet,
+   there is a risk that code called by [Uutil.showProgress] may include Lwt
+   code. For this reason only, it is better to include the fds in this
+   function in the fd cleanup scheme (done automatically by [openFile*] and
+   [closeFile*] functions). *)
 let copyContents fspathFrom pathFrom fspathTo pathTo fileKind fileLength ido =
   let use_id f = match ido with Some id -> f id | None -> () in
   let inFd = openFileIn fspathFrom pathFrom fileKind in
@@ -363,16 +433,16 @@ let copyContents fspathFrom pathFrom fspathTo pathTo fileKind fileLength ido =
               (fun l ->
                  use_id (fun id ->
 (* (Util.msg "Copied file %s (%d bytes)\n" (Path.toString pathFrom) l); *)
+                   if fileKind <> `RESS then Abort.checkAll ();
                    Uutil.showProgress id (Uutil.Filesize.ofInt l) "l"));
-            close_in inFd;
-            close_out outFd;
+            closeFileIn inFd;
+            closeFileOut outFd;
 (* ignore (Sys.command ("ls -l " ^ (Fspath.toString (Fspath.concat fspathTo pathTo)))) *)
          )
-         (fun () -> close_out_noerr outFd))
-    (fun () -> close_in_noerr inFd)
+         (fun () -> closeFileOutNoErr outFd))
+    (fun () -> closeFileInNoErr inFd)
 
-let localFile
-     fspathFrom pathFrom fspathTo pathTo realPathTo update desc ressLength ido =
+let localFileContents fspathFrom pathFrom fspathTo pathTo desc ressLength ido =
   Util.convertUnixErrorsToTransient
     "copying locally"
     (fun () ->
@@ -385,19 +455,26 @@ let localFile
         fspathFrom pathFrom fspathTo pathTo `DATA (Props.length desc) ido;
       if ressLength > Uutil.Filesize.zero then
         copyContents
-          fspathFrom pathFrom fspathTo pathTo `RESS ressLength ido;
-      setFileinfo fspathTo pathTo realPathTo update desc)
+          fspathFrom pathFrom fspathTo pathTo `RESS ressLength ido)
+
+let localFile
+     fspathFrom pathFrom fspathTo pathTo realPathTo update desc ressLength ido =
+  Util.convertUnixErrorsToTransient "copying locally" (fun () ->
+    localFileContents fspathFrom pathFrom fspathTo pathTo desc ressLength ido;
+    let (_, desc) = propsWithExtDataLocal fspathFrom (`Local pathFrom) desc in
+    setFileinfo fspathTo pathTo realPathTo update desc)
 
 (****)
 
-let tryCopyMovedFile fspathTo pathTo realPathTo update desc fp ress id =
-  if not (Prefs.read Xferhint.xferbycopying) then None else
+let tryCopyMovedFile connFrom fspathFrom pathFrom fspathTo pathTo realPathTo
+      update desc fp ress id =
+  if not (Prefs.read Xferhint.xferbycopying) then Lwt.return None else
   Util.convertUnixErrorsToTransient "tryCopyMovedFile" (fun() ->
     debug (fun () -> Util.msg "tryCopyMovedFile: -> %s /%s/\n"
       (Path.toString pathTo) (Os.fullfingerprint_to_string fp));
     match Xferhint.lookup fp with
       None ->
-        None
+        Lwt.return None
     | Some (candidateFspath, candidatePath, hintHandle) ->
         debug (fun () ->
           Util.msg
@@ -409,14 +486,15 @@ let tryCopyMovedFile fspathTo pathTo realPathTo update desc fp ress id =
              [true] is correct.  Otherwise, we don't expect to point
              to a symlink, and therefore we still get the correct
              result. *)
-          let info = Fileinfo.get true candidateFspath candidatePath in
+          let info = Fileinfo.getBasic true candidateFspath candidatePath in
           if
             info.Fileinfo.typ <> `ABSENT &&
             Props.length info.Fileinfo.desc = Props.length desc
           then begin
-            localFile
-              candidateFspath candidatePath fspathTo pathTo realPathTo
-              update desc (Osx.ressLength ress) (Some id);
+            localFileContents candidateFspath candidatePath fspathTo pathTo desc
+              (Osx.ressLength ress) (Some id);
+            propsWithExtData connFrom fspathFrom pathFrom desc >>= fun desc ->
+            setFileinfo fspathTo pathTo realPathTo update desc;
             let (info, isTransferred) =
               fileIsTransferred fspathTo pathTo desc fp ress in
             if isTransferred then begin
@@ -429,29 +507,29 @@ let tryCopyMovedFile fspathTo pathTo realPathTo update desc fp ress id =
                  (Fspath.toPrintString candidateFspath)
                  (Path.toString candidatePath)
               in
-              Some (info, msg)
+              Lwt.return (Some (info, msg))
             end else begin
               debug (fun () ->
                 Util.msg "tryCopyMoveFile: candidate file %s modified!\n"
                   (Path.toString candidatePath));
               Xferhint.deleteEntry hintHandle;
-              None
+              Lwt.return None
             end
           end else begin
             debug (fun () ->
               Util.msg "tryCopyMoveFile: candidate file %s disappeared!\n"
                 (Path.toString candidatePath));
             Xferhint.deleteEntry hintHandle;
-            None
+            Lwt.return None
           end
         with
           Util.Transient s ->
             debug (fun () ->
               Util.msg
-                "tryCopyMovedFile: local copy from %s didn't work [%s]"
+                "tryCopyMovedFile: local copy from %s didn't work [%s]\n"
                 (Path.toString candidatePath) s);
             Xferhint.deleteEntry hintHandle;
-            None)
+            Lwt.return None)
 
 (****)
 
@@ -473,11 +551,15 @@ let rsyncActivated =
 
 let decompressor = ref Remote.MsgIdMap.empty
 
+let resetDecompressorState () =
+  decompressor := Remote.MsgIdMap.empty
+let () = Remote.at_conn_close resetDecompressorState
+
 let processTransferInstruction conn (file_id, ti) =
   Util.convertUnixErrorsToTransient
     "processing a transfer instruction"
     (fun () ->
-       ignore (Remote.MsgIdMap.find file_id !decompressor ti))
+       ignore ((fst (Remote.MsgIdMap.find file_id !decompressor)) ti))
 
 let marshalTransferInstruction =
   (fun _ (file_id, (data, pos, len)) rem ->
@@ -511,6 +593,7 @@ let compress conn
               (fun () ->
                  showPrefixProgress id fileKind;
                  let showProgress count =
+                   if fileKind <> `RESS then Abort.checkAll ();
                    Uutil.showProgress id (Uutil.Filesize.ofInt count) "r" in
                  let compr =
                    match biOpt with
@@ -523,10 +606,10 @@ let compress conn
                  compr
                    (fun ti -> processTransferInstructionRemotely (file_id, ti))
                        >>= fun () ->
-                 close_in infd;
+                 closeFileIn infd;
                  Lwt.return ())
               (fun () ->
-                 close_in_noerr infd)))
+                 closeFileInNoErr infd)))
     (fun e ->
        (* We cannot wrap the code above with the handler below,
           as the code is executed asynchronously. *)
@@ -562,21 +645,21 @@ let close_all infd outfd =
     "closing files"
     (fun () ->
        begin match !infd with
-         Some fd -> close_in fd; infd := None
+         Some fd -> closeFileIn fd; infd := None
        | None    -> ()
        end;
        begin match !outfd with
-         Some fd -> close_out fd; outfd := None
+         Some fd -> closeFileOut fd; outfd := None
        | None    -> ()
        end)
 
 let close_all_no_error infd outfd =
   begin match !infd with
-    Some fd -> close_in_noerr fd
+    Some fd -> closeFileInNoErr fd
   | None    -> ()
   end;
   begin match !outfd with
-    Some fd -> close_out_noerr fd
+    Some fd -> closeFileOutNoErr fd
   | None    -> ()
   end
 
@@ -604,11 +687,12 @@ let referenceFd fspath path kind infd =
   | Some fd ->
       fd
 
-let rsyncReg = Lwt_util.make_region (40 * 1024)
+let rsyncReg = Remote.lwtRegionWithConnCleanup (40 * 1024)
+
 let rsyncThrottle useRsync srcFileSize destFileSize f =
   if not useRsync then f () else
   let l = Transfer.Rsync.memoryFootprint srcFileSize destFileSize in
-  Lwt_util.run_in_region rsyncReg l f
+  Lwt_util.run_in_region !rsyncReg l f
 
 let transferFileContents
       connFrom fspathFrom pathFrom fspathTo pathTo realPathTo update
@@ -619,6 +703,7 @@ let transferFileContents
   let outfd = ref None in
   let infd = ref None in
   let showProgress count =
+    if fileKind <> `RESS then Abort.checkAll ();
     Uutil.showProgress id (Uutil.Filesize.ofInt count) "r" in
 
   let destFileSize =
@@ -648,7 +733,7 @@ let transferFileContents
                protect
                  (fun () -> Transfer.Rsync.rsyncPreprocess
                               ifd srcFileSize destFileSize)
-                 (fun () -> close_in_noerr ifd)
+                 (fun () -> closeFileInNoErr ifd)
              in
              close_all infd outfd;
              (Some bi,
@@ -675,7 +760,7 @@ let transferFileContents
     Lwt.catch
       (fun () ->
          debug (fun () -> Util.msg "Starting the actual transfer\n");
-         decompressor := Remote.MsgIdMap.add file_id decompr !decompressor;
+         decompressor := Remote.MsgIdMap.add file_id (decompr, (infd, outfd)) !decompressor;
          compressRemotely connFrom
            ((bi, fspathFrom, pathFrom, fileKind), (srcFileSize, id, file_id))
            >>= fun () ->
@@ -708,6 +793,7 @@ let transferResourceForkAndSetFileinfo
   end else
     Lwt.return ()
   end >>= fun () ->
+  propsWithExtData connFrom fspathFrom pathFrom desc >>= fun desc ->
   setFileinfo fspathTo pathTo realPathTo update desc;
   debug (fun() -> Util.msg "Resource fork transferred for %s; doing last paranoid check\n"
     (Path.toString realPathTo));
@@ -744,6 +830,11 @@ let reallyTransferFile
 
 let filesBeingTransferred = Hashtbl.create 17
 
+let resetFileTransferState () =
+  (* The waiting threads should be collected by GC *)
+  Hashtbl.clear filesBeingTransferred
+let () = Remote.at_conn_close resetFileTransferState
+
 let wakeupNextTransfer fp =
   match
     try
@@ -820,17 +911,8 @@ let copythreshold =
      ^ "See \\sectionref{speeding}{Making Unison Faster on Large Files} "
      ^ "for more information.")
 
-let copyquoterem =
-  Prefs.createBoolWithDefault "copyquoterem"
-    ~category:(`Advanced `General)
-    "add quotes to remote file name for copyprog (true/false/default)"
-    ("When set to {\\tt true}, this flag causes Unison to add an extra layer "
-     ^ "of quotes to the remote path passed to the external copy program. "
-     ^ "This is needed by rsync, for example, which internally uses an ssh "
-     ^ "connection requiring an extra level of quoting for paths containing "
-     ^ "spaces. When this flag is set to {\\tt default}, extra quotes are "
-     ^ "added if the value of {\\tt copyprog} contains the string "
-     ^ "{\\tt rsync}.")
+(* Pref copyquoterem removed since 2.53.3 *)
+let () = Prefs.markRemoved "copyquoterem"
 
 let copymax =
   Prefs.createInt "copymax" 1
@@ -846,7 +928,7 @@ let formatConnectionInfo root =
       (* Find the (unique) nonlocal root *)
       match
          Safelist.find (function Clroot.ConnectLocal _ -> false | _ -> true)
-           (Safelist.map Clroot.parseRoot (Globals.rawRoots()))
+           (Globals.parsedClRawRoots ())
       with
         Clroot.ConnectByShell (_,rawhost,uo,_,_) ->
             (match uo with None -> "" | Some u -> u ^ "@")
@@ -870,7 +952,7 @@ let shouldUseExternalCopyprog update desc =
   && update = `Copy
 
 let prepareExternalTransfer fspathTo pathTo =
-  let info = Fileinfo.get false fspathTo pathTo in
+  let info = Fileinfo.getBasic false fspathTo pathTo in
   match info.Fileinfo.typ with
     `FILE when Props.length info.Fileinfo.desc > Uutil.Filesize.zero ->
       let perms = Props.perms info.Fileinfo.desc in
@@ -890,7 +972,7 @@ let prepareExternalTransfer fspathTo pathTo =
 let finishExternalTransferLocal connFrom
       ((fspathFrom, pathFrom, fspathTo, pathTo, realPathTo),
        (update, desc, fp, ress, id)) =
-  let info = Fileinfo.get false fspathTo pathTo in
+  let info = Fileinfo.getBasic false fspathTo pathTo in
   if
     info.Fileinfo.typ <> `FILE ||
     Props.length info.Fileinfo.desc <> Props.length desc
@@ -934,40 +1016,31 @@ let finishExternalTransferOnRoot =
     "finishExternalTransfer" ~convV0
     mfinishExternalTransfer mtransferStatus finishExternalTransferLocal
 
-let copyprogReg = Lwt_util.make_region 1
+let copyprogReg = Remote.lwtRegionWithConnCleanup 1
 
 let transferFileUsingExternalCopyprog
              rootFrom pathFrom rootTo fspathTo pathTo realPathTo
              update desc fp ress id useExistingTarget =
   Uutil.showProgress id Uutil.Filesize.zero "ext";
-  let prog =
+  let progWithArgs =
     if useExistingTarget then
       Prefs.read copyprogrest
     else
       Prefs.read copyprog
   in
-  let extraquotes = Prefs.read copyquoterem = `True
-                 || (  Prefs.read copyquoterem = `Default
-                    && Util.findsubstring "rsync" prog <> None) in
-  let addquotes root s =
-    match root with
-    | Common.Local, _ -> s
-    | Common.Remote _, _ -> if extraquotes then Uutil.quotes s else s in
   let fromSpec =
       (formatConnectionInfo rootFrom)
-    ^ (addquotes rootFrom
-         (Fspath.toString (Fspath.concat (snd rootFrom) pathFrom))) in
+    ^ (Fspath.toString (Fspath.concat (snd rootFrom) pathFrom)) in
   let toSpec =
       (formatConnectionInfo rootTo)
-    ^ (addquotes rootTo
-         (Fspath.toString (Fspath.concat fspathTo pathTo))) in
-  let cmd = prog ^ " "
-             ^ (Uutil.quotes fromSpec) ^ " "
-             ^ (Uutil.quotes toSpec) in
-  Trace.log (Printf.sprintf "%s\n" cmd);
-  Lwt_util.resize_region copyprogReg (Prefs.read copymax);
-  Lwt_util.run_in_region copyprogReg 1
-    (fun () -> External.runExternalProgram cmd) >>= fun (_, log) ->
+    ^ (Fspath.toString (Fspath.concat fspathTo pathTo)) in
+  Trace.log (progWithArgs ^ " " ^ fromSpec ^ " " ^ toSpec ^ "\n");
+  Lwt_util.resize_region !copyprogReg (Prefs.read copymax);
+  let args = Str.split (Str.regexp "[ \t]+") progWithArgs in
+  let prog = match args with [] -> assert false | h :: _ -> h in
+  Lwt_util.run_in_region !copyprogReg 1
+    (fun () -> External.runExternalProgramArgs prog
+                 (Array.of_list (args @ [fromSpec; toSpec]))) >>= fun (_, log) ->
   debug (fun() ->
            let l = Util.trimWhitespace log in
            Util.msg "transferFileUsingExternalCopyprog %s: returned...\n%s%s"
@@ -994,16 +1067,16 @@ let transferFileLocal connFrom
         (Fspath.toDebugString fspathTo) (Path.toString realPathTo) in
     let len = Uutil.Filesize.add (Props.length desc) (Osx.ressLength ress) in
     Uutil.showProgress id len "alr";
+    propsWithExtData connFrom fspathFrom pathFrom desc >>= fun desc ->
     setFileinfo fspathTo pathTo realPathTo update desc;
     Xferhint.insertEntry fspathTo pathTo fp;
     Lwt.return (`DONE (TransferSucceeded tempInfo, Some msg))
   end else
     registerFileTransfer pathTo fp
       (fun () ->
-         match
-           tryCopyMovedFile fspathTo pathTo realPathTo update desc fp ress id
-         with
-           Some (info, msg) ->
+         tryCopyMovedFile connFrom fspathFrom pathFrom
+           fspathTo pathTo realPathTo update desc fp ress id >>= function
+         | Some (info, msg) ->
              (* Transfer was performed by copying *)
              Xferhint.insertEntry fspathTo pathTo fp;
              Lwt.return (`DONE (TransferSucceeded info, Some msg))
@@ -1049,7 +1122,7 @@ let transferFileOnRoot =
 
 (* We limit the size of the output buffers to about 512 KB
    (we cannot go above the limit below plus 64) *)
-let transferFileReg = Lwt_util.make_region 440
+let transferFileReg = Remote.lwtRegionWithConnCleanup 440
 
 let bufferSize sz =
     (* Token queue *)
@@ -1095,7 +1168,7 @@ let transferFile
     f ()
   else
     let bufSz = bufferSize (max (Props.length desc) (Osx.ressLength ress)) in
-    Lwt_util.run_in_region transferFileReg bufSz f
+    Lwt_util.run_in_region !transferFileReg bufSz f
 
 (****)
 
@@ -1141,7 +1214,8 @@ let file rootFrom pathFrom rootTo fspathTo pathTo realPathTo
       checkForChangesToSource rootFrom pathFrom desc fp stamp ress None true
         >>= fun () ->
       (* This function never returns (it is supposed to fail) *)
-      saveTempFileOnRoot rootTo (pathTo, realPathTo, reason)
+      saveTempFileOnRoot rootTo (pathTo, realPathTo, reason) >>= fun () ->
+      assert false
 
 (****)
 
@@ -1165,7 +1239,7 @@ let recursively fspathFrom pathFrom fspathTo pathTo =
         debug (fun () -> Util.msg "  Copying directory %s / %s to %s / %s\n"
           (Fspath.toDebugString fspathFrom) (Path.toString pFrom)
           (Fspath.toDebugString fspathTo) (Path.toString pTo));
-        Os.createDir fspathTo pTo info.Fileinfo.desc;
+        Os.createDir fspathTo pTo (Props.perms info.Fileinfo.desc);
         let ch = Os.childrenOf fspathFrom pFrom in
         Safelist.iter
           (fun n -> copy (Path.child pFrom n) (Path.child pTo n)) ch
