@@ -46,20 +46,26 @@ let writeCommitLog source target tempname =
        Printf.fprintf c "(and delete this notice when you've done so).\n";
        close_out c)
 
-let clearCommitLog pathTo =
+let clearCommitLog tmpName =
   debug (fun() -> (Util.msg "Deleting commit log\n"));
 
-  let tmpPathDir = Fspath.canonize (Some Util.unisonDirStr) in  (* tmpPathDir is a Fspath.t *)
-  (* Use pathTo in the temporary name (instead of DANGER.README) to reduce chance of reuse *)
-  let tmpPath = Os.tempPath tmpPathDir pathTo in  (* tmpPath is a Path.local *)
-  let dangerFspath = Fspath.canonize (Some (System.fspathToString commitLogName)) in
-  let dangerFsPathTmp = Fspath.concat tmpPathDir tmpPath in
+  let commitLogNameWin () =
+    (* Work around an issue in Windows where unlink may not be immediate. *)
+    let p = commitLogName ^ (Filename.basename (Path.toString tmpName)) in
+    let rec tmp n =
+      let p = p ^ (string_of_int n) in
+      if System.file_exists p then tmp (n + 1)
+      else (System.rename commitLogName p; p)
+    in
+    try tmp 0 with
+    | Sys_error _ | Unix.Unix_error _ -> commitLogName
+  in
+  let commitLogUnlinkPath =
+    if Util.osType = `Win32 then commitLogNameWin () else commitLogName in
 
-  Os.renameFspath "DANGER.README" dangerFspath dangerFsPathTmp;
-  
   Util.convertUnixErrorsToFatal
     "clearing commit log"
-      (fun () -> System.unlink (System.fspathFromString (Fspath.toString dangerFsPathTmp)) )
+      (fun () -> System.unlink commitLogUnlinkPath)
 
 let processCommitLog () =
   if System.file_exists commitLogName then begin
@@ -68,7 +74,7 @@ let processCommitLog () =
             "Warning: the previous run of %s terminated in a dangerous state.
             Please consult the file %s, delete it, and try again."
                 Uutil.myName
-                (System.fspathToPrintString commitLogName)))
+                commitLogName))
   end else
     Lwt.return ()
 
@@ -77,7 +83,7 @@ let processCommitLogOnHost =
 
 let processCommitLogs() =
   Lwt_unix.run
-    (Globals.allHostsIter (fun h -> processCommitLogOnHost h ()))
+    (Globals.allRootsIter (fun r -> processCommitLogOnHost r ()))
 
 (* ------------------------------------------------------------ *)
 
@@ -93,7 +99,7 @@ let copyOnConflict = Prefs.createBool "copyonconflict" false
 
 let prepareCopy workingDir path notDefault =
   if notDefault && Prefs.read copyOnConflict then begin
-    match (Fileinfo.get true workingDir path).Fileinfo.typ with
+    match Fileinfo.getType true workingDir path with
     | `ABSENT -> Some (workingDir, path, None)
     | _ ->
       begin
@@ -171,6 +177,7 @@ let setPropLocal (fspath, (path, ui, newDesc, oldDesc)) =
   let localPath = Update.translatePathLocal fspath path in
   let (workingDir,realPath) = Fspath.findWorkingDir fspath localPath in
   Fileinfo.set workingDir realPath (`Update oldDesc) newDesc;
+  let newDesc = Props.purgeExtData newDesc in
   if fileUpdated ui then Stasher.stashCurrentVersion fspath localPath None;
   (* Archive update must be done last *)
   Update.updateProps fspath localPath (Some newDesc) ui;
@@ -185,7 +192,7 @@ let convV0 = Remote.makeConvV0FunArg
          Props.of_compat251 newDesc, Props.of_compat251 oldDesc)))
 
 let setPropOnRoot = Remote.registerRootCmd "setProp" ~convV0
-  Umarshal.(prod4 Path.m Common.mupdateItem Props.m Props.m id id) Umarshal.unit
+  Umarshal.(prod4 Path.m Common.mupdateItem Props.mx Props.m id id) Umarshal.unit
   setPropLocal
 
 let propOpt_to_compat251 = function
@@ -198,17 +205,21 @@ let propOpt_of_compat251 = function
 
 let convV0 = Remote.makeConvV0FunArg
   (fun (fspath, (path, propOpt, ui)) ->
-       (fspath, (path, propOpt_to_compat251 propOpt, Common.ui_to_compat251 ui)))
+       (fspath, (Path.makeGlobal path, propOpt_to_compat251 propOpt,
+         Common.ui_to_compat251 ui)))
   (fun (fspath, (path, propOpt, ui)) ->
-       (fspath, (path, propOpt_of_compat251 propOpt, Common.ui_of_compat251 ui)))
+       (fspath, (Path.forceLocal path,
+         propOpt_of_compat251 propOpt, Common.ui_of_compat251 ui)))
 
 let updatePropsOnRoot =
   Remote.registerRootCmd
    "updateProps" ~convV0
-   Umarshal.(prod3 Path.m (option Props.m) Common.mupdateItem id id)
+   Umarshal.(prod3 Path.mlocal (option Props.m) Common.mupdateItem id id)
    Umarshal.unit
      (fun (fspath, (path, propOpt, ui)) ->
-        let localPath = Update.translatePathLocal fspath path in
+        (* Previous versions of this function received a global path as input *)
+        let localPath = if Props.xattrEnabled () then path
+          else Update.translatePathLocal fspath (Path.makeGlobal path) in
         (* Archive update must be done first *)
         Update.updateProps fspath localPath propOpt ui;
         if fileUpdated ui then
@@ -228,8 +239,12 @@ let setProp rootFrom pathFrom rootTo pathTo newDesc oldDesc uiFrom uiTo =
       (Props.toString newDesc)
       (root2string rootTo) (Path.toString pathTo)
       (Props.toString oldDesc));
+  Copy.readPropsExtDataG rootFrom pathFrom newDesc >>= fun (p, newDesc) ->
   setPropOnRoot rootTo (pathTo, uiTo, newDesc, oldDesc) >>= fun _ ->
-  updateProps rootFrom pathFrom None uiFrom
+  (match p with
+  | None -> Update.translatePath rootFrom pathFrom
+  | Some path -> Lwt.return path) >>= fun localPathFrom ->
+  updateProps rootFrom localPathFrom None uiFrom
 
 (* ------------------------------------------------------------ *)
 
@@ -241,9 +256,9 @@ let mkdirOnRoot =
   Remote.registerRootCmd
     "mkdir" ~convV0
     Umarshal.(prod2 Fspath.m Path.mlocal id id)
-    Umarshal.(prod2 bool Props.m id id)
+    Umarshal.(prod2 bool Props.mbasic id id)
     (fun (fspath,(workingDir,path)) ->
-       let info = Fileinfo.get false workingDir path in
+       let info = Fileinfo.getBasic false workingDir path in
        if info.Fileinfo.typ = `DIRECTORY then begin
          begin try
            (* Make sure the directory is writable *)
@@ -254,8 +269,8 @@ let mkdirOnRoot =
        end else begin
          if info.Fileinfo.typ <> `ABSENT then
            Os.delete workingDir path;
-         Os.createDir workingDir path Props.dirDefault;
-         Lwt.return (false, (Fileinfo.get false workingDir path).Fileinfo.desc)
+         Os.createDir workingDir path (Props.perms Props.dirDefault);
+         Lwt.return (false, (Fileinfo.getBasic false workingDir path).desc)
        end)
 
 let convV0 = Remote.makeConvV0FunArg
@@ -269,7 +284,7 @@ let convV0 = Remote.makeConvV0FunArg
 let setDirPropOnRoot =
   Remote.registerRootCmd
     "setDirProp" ~convV0
-    Umarshal.(prod4 Fspath.m Path.mlocal Props.m Props.m id id)
+    Umarshal.(prod4 Fspath.m Path.mlocal Props.mbasic Props.mx id id)
     Umarshal.unit
     (fun (_, (workingDir, path, initialDesc, newDesc)) ->
       Fileinfo.set workingDir path (`Set initialDesc) newDesc;
@@ -301,15 +316,15 @@ let performRename fspathTo localPathTo workingDir pathFrom pathTo prevArch =
        (Fspath.toDebugString source) (Fspath.toDebugString target))
     (fun () ->
       debugverbose (fun() ->
-        Util.msg "calling Fileinfo.get from renameLocal\n");
+        Util.msg "calling Fileinfo.getType from renameLocal\n");
       let filetypeFrom =
-        (Fileinfo.get false source Path.empty).Fileinfo.typ in
+        Fileinfo.getType false source Path.empty in
       debugverbose (fun() ->
-        Util.msg "back from Fileinfo.get from renameLocal\n");
+        Util.msg "back from Fileinfo.getType from renameLocal\n");
       if filetypeFrom = `ABSENT then raise (Util.Transient (Printf.sprintf
            "Error while renaming %s to %s -- source file has disappeared!"
            (Fspath.toPrintString source) (Fspath.toPrintString target)));
-      let filetypeTo = (Fileinfo.get false target Path.empty).Fileinfo.typ in
+      let filetypeTo = Fileinfo.getType false target Path.empty in
 
        (* Windows and Unix operate differently if the target path of a
           rename already exists: in Windows an exception is raised, in
@@ -350,7 +365,7 @@ let performRename fspathTo localPathTo workingDir pathFrom pathTo prevArch =
                        (Fspath.toDebugString target));
               Os.rename "renameLocal(2)"
                 source Path.empty target Path.empty))
-          (fun _ -> clearCommitLog pathTo);
+          (fun _ -> clearCommitLog tmpPath);
         (* It is ok to leave a temporary file.  So, the log can be
            cleared before deleting it. *)
         Os.delete temp Path.empty
@@ -502,10 +517,19 @@ let convV0 = Remote.makeConvV0FunArg
 
 let setupTargetPathsAndCreateParentDirectory =
   Remote.registerRootCmd "setupTargetPathsAndCreateParentDirectory" ~convV0
-    Umarshal.(prod2 Path.m (list Props.m) id id)
+    Umarshal.(prod2 Path.m (list Props.mx) id id)
     Umarshal.(prod4 Fspath.m Path.mlocal Path.mlocal Path.mlocal id id)
     setupTargetPathsAndCreateParentDirectoryLocal
 
+let rec readParentsExtData rootFrom pathFrom acc = function
+  | [] -> Safelist.rev acc |> Lwt.return
+  | desc :: rem ->
+      match Path.deconstructRev pathFrom with
+      | None -> assert false
+      | Some (_, parentPath) ->
+          Copy.readPropsExtData rootFrom parentPath desc >>= fun desc' ->
+          readParentsExtData rootFrom parentPath (desc' :: acc) rem
+
 (* ------------------------------------------------------------ *)
 
 let updateSourceArchiveLocal (fspathFrom, (localPathFrom, uiFrom, errPaths)) =
@@ -579,7 +603,7 @@ let rec normalizeProps propsFrom propsTo =
 
 (* ------------------------------------------------------------ *)
 
-let copyReg = Lwt_util.make_region 50
+let copyReg = Remote.lwtRegionWithConnCleanup 50
 
 let copy
       update
@@ -600,9 +624,13 @@ let copy
       "copy %s %s ---> %s %s \n"
       (root2string rootFrom) (Path.toString pathFrom)
       (root2string rootTo) (Path.toString pathTo));
+  (* Calculate source path *)
+  Update.translatePath rootFrom pathFrom >>= fun localPathFrom ->
   (* Calculate target paths *)
+  normalizeProps propsFrom propsTo
+  |> readParentsExtData rootFrom localPathFrom [] >>= fun parentProps ->
   setupTargetPathsAndCreateParentDirectory rootTo
-    (pathTo, normalizeProps propsFrom propsTo)
+    (pathTo, parentProps)
      >>= fun (workingDir, realPathTo, tempPathTo, localPathTo) ->
   (* When in Unicode case-insensitive mode, we want to create files
      with NFC normal-form filenames. *)
@@ -617,8 +645,6 @@ let copy
         | Some (name, parentPath) ->
             Path.child parentPath (Name.normalize name)
   in
-  (* Calculate source path *)
-  Update.translatePath rootFrom pathFrom >>= fun localPathFrom ->
   let errors = ref [] in
   (* Inner loop for recursive copy... *)
   let rec copyRec pFrom      (* Path to copy from *)
@@ -635,7 +661,7 @@ let copy
       (fun () ->
          match f with
            Update.ArchiveFile (desc, fp, stamp, ress) ->
-             Lwt_util.run_in_region copyReg 1 (fun () ->
+             Lwt_util.run_in_region !copyReg 1 (fun () ->
                Abort.check id;
                let stmp =
                  if Update.useFastChecking () then Some stamp else None in
@@ -649,14 +675,14 @@ let copy
                                       fp, Fileinfo.stamp info, ress'),
                   []))
          | Update.ArchiveSymlink l ->
-             Lwt_util.run_in_region copyReg 1 (fun () ->
+             Lwt_util.run_in_region !copyReg 1 (fun () ->
                debug (fun() -> Util.msg "Making symlink %s/%s -> %s\n"
                                  (root2string rootTo) (Path.toString pTo) l);
                Abort.check id;
                makeSymlink rootTo (workingDir, pTo, l) >>= fun () ->
                Lwt.return (f, []))
          | Update.ArchiveDir (desc, children) ->
-             Lwt_util.run_in_region copyReg 1 (fun () ->
+             Lwt_util.run_in_region !copyReg 1 (fun () ->
                debug (fun() -> Util.msg "Creating directory %s/%s\n"
                  (root2string rootTo) (Path.toString pTo));
                mkdirOnRoot rootTo (workingDir, pTo))
@@ -694,11 +720,12 @@ let copy
              else
                Lwt.return ()
              end >>= fun () ->
-             Lwt_util.run_in_region copyReg 1 (fun () ->
+             Copy.readPropsExtData rootFrom pFrom desc >>= fun desc' ->
+             Lwt_util.run_in_region !copyReg 1 (fun () ->
                (* We use the actual file permissions so as to preserve
                   inherited bits *)
                setDirPropOnRoot rootTo
-                 (workingDir, pTo, initialDesc, desc)) >>= fun () ->
+                 (workingDir, pTo, initialDesc, desc')) >>= fun () ->
              Lwt.return (Update.ArchiveDir (desc, newChildren),
                          List.flatten pathl)
          | Update.NoArchive ->
@@ -706,10 +733,8 @@ let copy
       (fun e ->
          match e with
            Util.Transient _ ->
-             if not (Abort.testException e) then begin
-               Abort.file id;
-               errors := e :: !errors
-             end;
+             if not (Abort.testException e) then Abort.file id;
+             errors := e :: !errors;
              Lwt.return (Update.NoArchive, [pFrom])
          | _ ->
              Lwt.fail e)
@@ -914,8 +939,8 @@ let formatMergeCmd p f1 f2 backup out1 out2 outarch batchmode =
 let copyBack fspathFrom pathFrom rootTo pathTo propsTo uiTo archTo id =
   setupTargetPaths rootTo pathTo
     >>= (fun (workingDirForCopy, realPathTo, tempPathTo, localPathTo) ->
-  let info = Fileinfo.get false fspathFrom pathFrom in
-  let fp = Os.fingerprint fspathFrom pathFrom info in
+  let info = Fileinfo.getBasicWithRess false fspathFrom pathFrom in
+  let fp = Os.fingerprint fspathFrom pathFrom info.Fileinfo.typ in
   let stamp = Osx.stamp info.Fileinfo.osX in
   let newprops = Props.setLength propsTo (Props.length info.Fileinfo.desc) in
   Copy.file
@@ -1036,10 +1061,10 @@ let merge root1 path1 ui1 root2 path2 ui2 id showMergeFn =
       Os.delete workingDirForMerge new1;
       Os.delete workingDirForMerge new2;
       Os.delete workingDirForMerge newarch;
-      let info1 = Fileinfo.get false workingDirForMerge working1 in
+      let info1 = Fileinfo.getType false workingDirForMerge working1 in
       (* FIX: Why split out the parts of the pair?  Why is it not abstract anyway??? *)
       let fp1 = Os.fingerprint workingDirForMerge working1 info1 in
-      let info2 = Fileinfo.get false workingDirForMerge working2 in
+      let info2 = Fileinfo.getType false workingDirForMerge working2 in
       let fp2 = Os.fingerprint workingDirForMerge working2 info2 in
       let cmd = formatMergeCmd
           path1
@@ -1089,8 +1114,8 @@ let merge root1 path1 ui1 root2 path2 ui2 id showMergeFn =
           say (fun () -> Util.msg "Three outputs detected \n")
         else
           say (fun () -> Util.msg "Two outputs detected \n");
-        let info1 = Fileinfo.get false workingDirForMerge new1 in
-        let info2 = Fileinfo.get false workingDirForMerge new2 in
+        let info1 = Fileinfo.getType false workingDirForMerge new1 in
+        let info2 = Fileinfo.getType false workingDirForMerge new2 in
         let fp1' = Os.fingerprint workingDirForMerge new1 info1 in
         let fp2' = Os.fingerprint workingDirForMerge new2 info2 in
         if fp1'=fp2' then begin
@@ -1130,9 +1155,9 @@ let merge root1 path1 ui1 root2 path2 ui2 id showMergeFn =
 
         if working1_still_exists && working2_still_exists then begin
           say (fun () -> Util.msg "No output from merge cmd and both original files are still present\n");
-          let info1' = Fileinfo.get false workingDirForMerge working1 in
+          let info1' = Fileinfo.getType false workingDirForMerge working1 in
           let fp1' = Os.fingerprint workingDirForMerge working1 info1' in
-          let info2' = Fileinfo.get false workingDirForMerge working2 in
+          let info2' = Fileinfo.getType false workingDirForMerge working2 in
           let fp2' = Os.fingerprint workingDirForMerge working2 info2' in
           if fp1 = fp1' && fp2 = fp2' then
             raise (Util.Transient "Merge program didn't change either temp file");
@@ -1194,17 +1219,50 @@ let merge root1 path1 ui1 root2 path2 ui2 id showMergeFn =
                    (Path.toString path1));
            if not (Stasher.shouldBackupCurrent path1) then
              Util.msg "Warning: 'backupcurrent' is not set for path %s\n" (Path.toString path1);
-           let infoarch = Fileinfo.get true arch_fspath Path.empty in
-           let fp = Os.fingerprint arch_fspath Path.empty infoarch in
+           let infoarch = Fileinfo.getBasicWithRess false arch_fspath Path.empty in
+           let fp = Os.fingerprint arch_fspath Path.empty infoarch.typ in
            debug (fun () -> Util.msg "New fingerprint is %s\n" (Os.fullfingerprint_to_string fp));
+           let pseudoMergeDesc merge_desc =
+             (* Length and times (because the merge result's mtime is set in
+                both replicas) must come from the merge result. The remaining
+                props should be as close as possible to one of the original
+                files to reduce the possibility of props conflicts at the next
+                sync.
+
+                Current props, desc1 and desc2, can't be compared before having
+                same time and length (taken from the merge result). *)
+             let fixup_desc desc n =
+               let desc' = Props.setTime desc n in
+               Props.setLength desc' (Props.length n)
+             in
+             let desc1' = fixup_desc desc1 merge_desc
+             and desc2' = fixup_desc desc2 merge_desc in
+             let pref_desc =
+               if Props.similar desc1' desc2' then Some desc1 else
+               match ui1, ui2 with
+               | Updates (_, Previous (_, pdesc1, _, _)),
+                 Updates (_, Previous (_, pdesc2, _, _)) ->
+                   if Props.similar pdesc1 desc1 then Some desc1 else
+                   if Props.similar pdesc2 desc2 then Some desc2 else
+                   if Props.similar pdesc1 pdesc2 then Some pdesc1 else
+                   None (* Is it possible to arrive here? *)
+               | NoUpdates, (NoUpdates | Updates _) -> Some desc1
+               | Updates _, NoUpdates -> Some desc2
+               | _ -> None
+             in
+             match pref_desc with
+             | None -> None
+             | Some pref_desc -> Some (fixup_desc pref_desc merge_desc)
+           in
            let new_archive_entry =
-             Update.ArchiveFile
-               (infoarch.desc, fp,
-                Fileinfo.stamp infoarch,
-                Osx.stamp infoarch.osX) in
-           (Props.setTime desc1 (Props.time infoarch.Fileinfo.desc),
-            Props.setTime desc2 (Props.time infoarch.Fileinfo.desc),
-            Some new_archive_entry)
+             match pseudoMergeDesc infoarch.desc with
+             | None -> None
+             | Some new_arch_desc ->
+                 Some (Update.ArchiveFile (new_arch_desc, fp,
+                   Fileinfo.stamp infoarch, Osx.stamp infoarch.osX)) in
+           (Props.setTime desc1 infoarch.Fileinfo.desc,
+            Props.setTime desc2 infoarch.Fileinfo.desc,
+            new_archive_entry)
          end else
            (desc1, desc2, None)
          in
