@@ -64,7 +64,7 @@ Protocol description
   only further changes.
 
   Unison can wait for changes in a replica by emitting a 'WAIT hash'
-  command. It can watch several replicas by sending a serie of these
+  command. It can watch several replicas by sending a series of these
   commands. The child process is expected to respond once, by a
   'CHANGE hash1 ... hash2' response that lists the changed replicas
   among those included in a 'WAIT' command, when changes are
@@ -222,26 +222,24 @@ let read_line i =
 (****)
 
 let path =
-  List.map System.fspathFromString
-    (try
+    try
        Str.split (Str.regexp (if Util.osType = `Win32 then ";" else ":"))
          (Sys.getenv "PATH")
      with Not_found ->
-       [])
+       []
 
 let search_in_path ?(path = path) name =
-  System.fspathConcat
+  Filename.concat
     (List.find (fun dir ->
-       let p = System.fspathConcat dir name in
+       let p = Filename.concat dir name in
        let found = System.file_exists p in
-       debug (fun () -> Util.msg "'%s' ...%s\n"
-         (System.fspathToString p)
+       debug (fun () -> Util.msg "'%s' ...%s\n" p
          (match found with true -> "found" | false -> "not found"));
        found)
     path)
     name
 
-let exec_path = [System.fspathFromString Sys.executable_name]
+let exec_path = [Sys.executable_name]
 (*
   try
     (* Linux *)
@@ -259,13 +257,12 @@ let exec_path = [System.fspathFromString Sys.executable_name]
       [System.fspathConcat (System.getcwd ()) name]
 *)
 
-let exec_dir = List.map System.fspathDirname exec_path
+let exec_dir = List.map Filename.dirname exec_path
 
 let watcher =
   lazy
     (let suffix = if Util.osType = `Win32 then ".exe" else "" in
      debug (fun () -> Util.msg "File monitoring helper program...\n");
-     System.fspathToString
        (try
           search_in_path ~path:(exec_dir @ path)
             ("unison-fsmonitor-" ^ Uutil.myMajorVersion ^ suffix)
@@ -277,6 +274,7 @@ type 'a exn_option = Value of 'a | Exn of exn | Nothing
 
 type conn =
   { output : Lwt_unix.file_descr;
+    pid : int;
     has_changes : Cond.t;
     has_line : Cond.t;
     line_read : Cond.t;
@@ -296,6 +294,9 @@ let rec reader conn read_line =
     reader conn read_line
    end
 
+let safeTerm pid =
+  try ignore (Terminal.safe_waitpid pid) with Unix.Unix_error _ -> ()
+
 let safeClose fd = try Lwt_unix.close fd with Unix.Unix_error _ -> ()
 
 let currentConnection () =
@@ -305,7 +306,7 @@ let currentConnection () =
 
 let closeConnection () =
   match !conn with
-    Some c -> conn := None; safeClose c.output
+  | Some c -> conn := None; safeClose c.output; safeTerm c.pid
   | None   -> ()
 
 let connected () = !conn <> None
@@ -313,15 +314,14 @@ let connected () = !conn <> None
 let startProcess () =
   try
     let w = Lazy.force watcher in
-    let (i1,o1) = Lwt_unix.pipe_out () in
-    let (i2,o2) = Lwt_unix.pipe_in () in
-    Lwt_unix.set_close_on_exec i2;
-    Lwt_unix.set_close_on_exec o1;
-    Util.convertUnixErrorsToFatal "starting filesystem watcher" (fun () ->
-      ignore (System.create_process w [|w|] i1 o2 Unix.stderr));
+    let (i1,o1) = Lwt_unix.pipe_out ~cloexec:true () in
+    let (i2,o2) = Lwt_unix.pipe_in ~cloexec:true () in
+    let pid = Util.convertUnixErrorsToFatal "starting filesystem watcher"
+      (fun () -> System.create_process w [|w|] i1 o2 Unix.stderr) in
     Unix.close i1; Unix.close o2;
     let c =
       { output = o1;
+        pid;
         has_changes = Cond.make ();
         has_line = Cond.make ();
         line_read = Cond.make ();
@@ -444,6 +444,8 @@ let start hash =
     true
   end
 
+let running _ = connected ()
+
 let wait hash =
   let c = currentConnection () in
   let res = Cond.wait c.has_changes in
