@@ -45,24 +45,36 @@ let setRawRoots l = Prefs.set rawroots (Safelist.rev l)
 
 let rawRoots () = Safelist.rev (Prefs.read rawroots)
 
+let parsedClrootCache = ref []
+
+let parsedClRawRoots () =
+  let key = Prefs.read rawroots in
+  match List.assq_opt key !parsedClrootCache with
+  | Some x -> x
+  | None -> let x = Safelist.map Clroot.parseRoot (rawRoots ()) in
+            parsedClrootCache := (key, x) :: !parsedClrootCache; x
+
+let wrongNumRootsExn roots =
+  Util.Fatal (Printf.sprintf "Wrong number of roots: \
+    2 expected, but %d provided (%s)\n(Maybe you specified \
+    roots both on the command line and in the profile?)"
+    (Safelist.length roots)
+    (String.concat ", " roots))
+
 let rawRootPair () =
   match rawRoots () with
     [r1; r2] -> (r1, r2)
-  | _        -> assert false
+  | roots    -> raise (wrongNumRootsExn roots)
 
 let theroots = ref []
 
-let uninstallRoots () = theroots := []
+let uninstallRoots () = theroots := []; parsedClrootCache := []
 
 open Lwt
 let installRoots termInteract =
   let () = uninstallRoots () in (* Clear out potential old roots *)
   let roots = rawRoots () in
-  if Safelist.length roots <> 2 then
-    raise (Util.Fatal (Printf.sprintf
-      "Wrong number of roots: 2 expected, but %d provided (%s)\n(Maybe you specified roots both on the command line and in the profile?)"
-      (Safelist.length roots)
-      (String.concat ", " roots) ));
+  if Safelist.length roots <> 2 then raise (wrongNumRootsExn roots);
   Safelist.fold_right
     (fun r cont ->
        Remote.canonizeRoot r (Clroot.parseRoot r) termInteract
@@ -143,25 +155,6 @@ let allRootsMapWithWaitingAction f wa =
     (rootsInCanonicalOrder ()) >>= (fun l ->
       return (Safelist.map snd (reorderCanonicalListToUsersOrder l)))
 
-let replicaHostnames () =
-  Safelist.map
-    (function (Local, _) -> ""
-            | (Remote h,_) -> h)
-    (rootsList())
-
-let allHostsIter f =
-  let rec iter l =
-    match l with
-      [] ->
-        return ()
-    | root :: rem ->
-        f root >>= (fun () ->
-        iter rem)
-  in
-  iter (replicaHostnames ())
-
-let allHostsMap f = Safelist.map f (replicaHostnames())
-
 let paths =
   Prefs.create "path" []
     ~category:(`Basic `Sync)
@@ -218,7 +211,7 @@ let propagatePrefs () =
   let toRoot = function
     | (Local, _) -> return ()
     | (Remote _, _) as root ->
-        let rpcVer = Remote.(connectionVersion (connectionToRoot root)) in
+        let rpcVer = Remote.(connectionVersion (connectionOfRoot root)) in
         let prefs = Prefs.dump rpcVer in
         propagatePrefsTo root root prefs
   in
