@@ -10,8 +10,11 @@ val rawRoots : unit -> string list
 val setRawRoots : string list -> unit
 val rawRootPair : unit -> string * string
 
+(* Same as [rawRoots], parsed as clroots                                     *)
+val parsedClRawRoots : unit -> Clroot.clroot list
+
 (* Parse and canonize roots from their raw names                             *)
-val installRoots : (string -> string -> string) option -> unit Lwt.t
+val installRoots : (string -> Terminal.termInteract) option -> unit Lwt.t
 
 (* An alternate method (under development?) *)
 val installRoots2 : unit -> unit
@@ -58,12 +61,6 @@ val paths : Path.t list Prefs.t
 (* Expand any paths ending with *                                            *)
 val expandWildcardPaths : unit -> unit
 
-(* Run a command on all hosts in roots                                       *)
-val allHostsIter : (string -> unit Lwt.t) -> unit Lwt.t
-
-(* Run a command on all hosts in roots and collect results                   *)
-val allHostsMap : (string -> 'a) -> 'a list
-
 (* Make sure that the server has the same settings for its preferences as we
    do locally.  Should be called whenever the local preferences have
    changed.  (This isn't conceptually a part of this module, but it can't
