@@ -215,9 +215,9 @@ let minfo = Umarshal.(prod2 (mressInfo (prod2 Fspath.m int64 id id)) string
                         (fun (ressInfo, finfo) -> {ressInfo; finfo}))
 
 external getFileInfosInternal :
-  System.fspath -> bool -> string * int64 = "getFileInfos"
+  string -> bool -> string * int64 = "getFileInfos"
 external setFileInfosInternal :
-  System.fspath -> string -> unit = "setFileInfos"
+  string -> string -> unit = "setFileInfos"
 
 let defaultInfos typ =
   match typ with
@@ -263,7 +263,7 @@ let getFileInfos dataFspath dataPath typ =
         try
           let (fInfo, rsrcLength) =
             getFileInfosInternal
-              (Fspath.toSysPath (Fspath.concat dataFspath dataPath))
+              (Fspath.toString (Fspath.concat dataFspath dataPath))
               (typ = `FILE)
           in
           { ressInfo =
@@ -274,7 +274,8 @@ let getFileInfos dataFspath dataPath typ =
           (* Not a HFS volume.  Look for an AppleDouble file *)
           try
             let (workingDir, realPath) =
-              Fspath.findWorkingDir dataFspath dataPath in
+              try Fspath.findWorkingDir dataFspath dataPath with
+              | Util.Transient _ -> raise Not_found in
             let (doubleFspath, inch, entries) =
               openDouble workingDir realPath in
             let (rsrcOffset, rsrcLength) =
@@ -361,18 +362,21 @@ let setFileInfos dataFspath dataPath finfo =
   assert (finfo <> "");
   Util.convertUnixErrorsToTransient "setting file information" (fun () ->
     try
-      let p = Fspath.toSysPath (Fspath.concat dataFspath dataPath) in
+      let p = Fspath.toString (Fspath.concat dataFspath dataPath) in
       let (fullFinfo, _) = getFileInfosInternal p false in
       setFileInfosInternal p (insertInfo (Bytes.of_string fullFinfo) finfo)
-    with Unix.Unix_error ((Unix.EOPNOTSUPP | Unix.ENOSYS), _, _) ->
+    with Unix.Unix_error ((EOPNOTSUPP | ENOSYS | EUNKNOWNERR 93), _, _) ->
+      (* ENOATTR (93) is returned on msdos/exfat fs since macOS 13 *)
       (* Not an HFS volume.  Look for an AppleDouble file *)
       let (workingDir, realPath) = Fspath.findWorkingDir dataFspath dataPath in
       begin try
         let (doubleFspath, inch, entries) = openDouble workingDir realPath in
         begin try
           let (ofs, len) = Safelist.assoc `FINFO entries in
-          if len < finfoLength then
-            fail dataFspath dataPath doubleFspath "bad finder info";
+          if len < finfoLength then begin
+            close_in_noerr inch;
+            fail dataFspath dataPath doubleFspath "bad finder info"
+          end;
           let fullFinfo =
             protect
               (fun () ->
@@ -471,7 +475,12 @@ let stamp info =
   | AppleDoubleRess (inode, mtime, ctime, len, _) ->
       AppleDoubleRess (inode, mtime, ctime, len, ())
 
-let ressFingerprint fspath path info =
+let ressFingerprint fspath path typ =
+  (* This function used to get ready-made info passed in. (Re-)getting the
+     info here may consume one or a few additional syscalls. This is not
+     thought to be a problem unless there are hundreds of thousands of files
+     with resource forks. That is really unlikely. *)
+  let info = getFileInfos fspath path typ in
   match info.ressInfo with
     NoRess ->
       Fingerprint.dummy
@@ -500,7 +509,7 @@ let openRessIn fspath path =
       Unix.in_channel_of_descr
         (Fs.openfile
            (Fspath.concat fspath (ressPath path))
-           [Unix.O_RDONLY] 0o444)
+           [Unix.O_RDONLY; O_CLOEXEC] 0o444)
     with Unix.Unix_error ((Unix.ENOENT | Unix.ENOTDIR), _, _) ->
       let (doublePath, inch, entries) = openDouble fspath path in
       try
@@ -518,7 +527,7 @@ let openRessOut fspath path length =
       let p = Fspath.concat fspath (ressPath path) in
       debug (fun () -> Util.msg "openRessOut %s\n" (Fspath.toString p));
       Unix.out_channel_of_descr
-        (Fs.openfile p [Unix.O_WRONLY;Unix.O_CREAT] 0o600)
+        (Fs.openfile p [Unix.O_WRONLY; O_CREAT; O_CLOEXEC] 0o600)
     with Unix.Unix_error ((Unix.ENOENT | Unix.ENOTDIR), _, _) ->
       debug (fun () -> Util.msg "Opening AppleDouble file for resource fork\n");
       let path = Fspath.appleDouble (Fspath.concat fspath path) in
