@@ -19,14 +19,9 @@ type fspath = string
 
 let mfspath = Umarshal.string
 
-let fspathFromString f = f
-let fspathToPrintString f = f
-let fspathToString f = f
-let fspathToDebugString f = String.escaped f
+let extendedPath f = f
 
-let fspathConcat = Filename.concat
-let fspathDirname = Filename.dirname
-let fspathAddSuffixToFinalName f suffix = f ^ suffix
+let fspathToDebugString f = String.escaped f
 
 (****)
 
@@ -73,8 +68,13 @@ let open_in_bin = open_in_bin
 
 let create_process = Unix.create_process
 let open_process_in = Unix.open_process_in
+let open_process_args_in = Unix.open_process_args_in
 let open_process_out = Unix.open_process_out
 let open_process_full cmd = Unix.open_process_full cmd (Unix.environment ())
+let open_process_args_full cmd args = Unix.open_process_args_full cmd args (Unix.environment ())
+let process_in_pid = Unix.process_in_pid
+let process_out_pid = Unix.process_out_pid
+let process_full_pid = Unix.process_full_pid
 let close_process_in = Unix.close_process_in
 let close_process_out = Unix.close_process_out
 let close_process_full = Unix.close_process_full
@@ -83,17 +83,13 @@ let close_process_full = Unix.close_process_full
 
 let isNotWindows = Sys.os_type <> "Win32"
 
-let canSetTime f =
-  isNotWindows ||
-  try
-    Unix.access f [Unix.W_OK];
-    true
-  with
-    Unix.Unix_error _ -> false
-
 (* Note that Cygwin provides some kind of inode numbers, but we only
    have access to the lower 32 bits on 32bit systems... *)
-let hasInodeNumbers () = isNotWindows
+(* Best effort inode numbers are provided in Windows since OCaml 4.03 *)
+(* However, these inode numbers are not usable on FAT filesystems, as
+   renaming a file "b" over a file "a" does not change the inode
+   number of "a". *)
+let hasInodeNumbers () = true
 
 let hasSymlink = Unix.has_symlink
 
@@ -124,10 +120,49 @@ let terminalStateFunctions () =
     startReading = (fun () -> ());
     stopReading = (fun () -> ()) }
 
+let termVtCapable fd = Unix.isatty fd
+
+let has_stdout ~info:_ = true
+let has_stderr ~info:_ = true
+
 (****)
 
 let fingerprint f =
   let ic = open_in_bin f in
-  let d = Digest.channel ic (-1) in
-  close_in ic;
-  d
+  try
+    let d = Digest.channel ic (-1) in
+    close_in ic;
+    d
+  with e ->
+    close_in_noerr ic;
+    raise e
+
+(****)
+
+exception XattrNotSupported
+let _ = Callback.register_exception "XattrNotSupported" XattrNotSupported
+
+external xattr_list : string -> (string * int) list = "unison_xattrs_list"
+external xattr_get_ : string -> string -> string = "unison_xattr_get"
+external xattr_set_ : string -> string -> string -> unit = "unison_xattr_set"
+external xattr_remove_ : string -> string -> unit = "unison_xattr_remove"
+external xattr_updates_ctime : unit -> bool = "unison_xattr_updates_ctime"
+
+let xattrUpdatesCTime = xattr_updates_ctime ()
+
+let xattr_get p n =
+  try xattr_get_ p n with
+  | Failure e -> failwith ("(attr: " ^ n ^ ") " ^ e)
+
+let xattr_set p n v =
+  try xattr_set_ p n v with
+  | Failure e -> failwith ("(attr: " ^ n ^ ") " ^ e)
+
+let xattr_remove p n =
+  try xattr_remove_ p n with
+  | Failure e -> failwith ("(attr: " ^ n ^ ") " ^ e)
+
+(****)
+
+external acl_get_text : string -> string = "unison_acl_to_text"
+external acl_set_text : string -> string -> unit = "unison_acl_from_text"
