@@ -45,7 +45,6 @@ val fingerprint : fspath -> Digest.t
 
 (****)
 
-val canSetTime : fspath -> bool
 val hasInodeNumbers : unit -> bool
 val hasSymlink : unit -> bool
 
@@ -56,23 +55,39 @@ val hasSymlink : unit -> bool
  * [hasCorrectCTime] can have a different value on different systems. *)
 val hasCorrectCTime : bool
 
+(****)
+
+exception XattrNotSupported
+
+val xattr_list : fspath -> (string * int) list
+val xattr_get : fspath -> string -> string
+val xattr_set : fspath -> string -> string -> unit
+val xattr_remove : fspath -> string -> unit
+
+(* [xattrUpdatesCTime] is true if changes to extended attributes update the
+ * file ctime. This means that extended attribute changes can be quickly
+ * detected by looking at ctime change. If file ctime is not updated then
+ * xattrs have to be scanned every time to detect changes. *)
+val xattrUpdatesCTime : bool
+
+(****)
+
+val acl_get_text : fspath -> string
+val acl_set_text : fspath -> string -> unit
+
 end
 
 module type Full = sig
 
-include Core
+include Core with type fspath = string
+
+val extendedPath : string -> fspath
 
 val putenv : string -> string -> unit
 val getenv : string -> string
 val argv : unit -> string array
 
-val fspathFromString : string -> fspath
-val fspathToPrintString : fspath -> string
 val fspathToDebugString : fspath -> string
-val fspathToString : fspath -> string
-val fspathConcat : fspath -> string -> fspath
-val fspathDirname : fspath -> fspath
-val fspathAddSuffixToFinalName : fspath -> string -> fspath
 
 val open_in_gen : open_flag list -> int -> fspath -> in_channel
 
@@ -84,9 +99,15 @@ val create_process :
   string -> string array ->
   Unix.file_descr -> Unix.file_descr -> Unix.file_descr -> int
 val open_process_in : string -> in_channel
+val open_process_args_in : string -> string array -> in_channel
 val open_process_out : string -> out_channel
 val open_process_full :
   string -> in_channel * out_channel * in_channel
+val open_process_args_full :
+  string -> string array -> in_channel * out_channel * in_channel
+val process_in_pid : in_channel -> int
+val process_out_pid : out_channel -> int
+val process_full_pid : in_channel * out_channel * in_channel -> int
 val close_process_in : in_channel -> Unix.process_status
 val close_process_out : out_channel -> Unix.process_status
 val close_process_full :
@@ -97,4 +118,9 @@ type terminalStateFunctions =
     startReading : unit -> unit; stopReading : unit -> unit }
 val terminalStateFunctions : unit -> terminalStateFunctions
 
+val termVtCapable : Unix.file_descr -> bool
+
+val has_stdout : info:string -> bool
+val has_stderr : info:string -> bool
+
 end
