@@ -235,12 +235,9 @@ let sync ?(verbose=false) () =
     displayRis reconItemList
   end;
   minisleep 0.1;
-  Lwt_unix.run (
-    Lwt_util.iter
-      (fun ri ->
-         Transport.transportItem ri
-           (Uutil.File.ofLine 0) (fun _ _ -> true))
-      reconItemList);
+  Uicommon.transportItems (Array.of_list reconItemList) (fun _ -> true)
+    (fun _ ri ->
+      Transport.transportItem ri (Uutil.File.ofLine 0) (fun _ _ -> true));
   Update.commitUpdates()
 
 let currentTest = ref ""
@@ -338,6 +335,22 @@ let test() =
      running fast enough that the whole thing happens within a second, then the
      update will be missed! *)
 
+  (* Test that update propagation transport works *)
+  let maxth = [| "0"; "1"; "5"; "6"; "7" |] in
+    (* Number of threads: default (0); 1 (corner case);
+       one less, equal to, and one more than number of updates *)
+  for i = 1 to Array.length maxth do
+    runtest ("propagation 1." ^ string_of_int i) ["maxthreads = " ^ maxth.(i - 1)] (fun () ->
+      put R1 (Dir []); put R2 (Dir []); sync ();
+      let r1 = ["a", File "a"; "b", File "b"; "d1", Dir ["a", File "a1"; "b", File "b1"]]
+      and r2 = ["x", File "x"; "y", File "y"; "d2", Dir ["x", File "x2"; "y", File "y2"]] in
+      let expect = Dir (r1 @ r2) in
+      put R1 (Dir r1); put R2 (Dir r2); sync ();
+      check "1" R1 expect;
+      check "2" R2 expect
+    )
+  done;
+
   (* Test that .git is treated atomically. *)
   runtest "Atomicity of certain directories 1" ["atomic = Name .git";
                                                 "force = newer"] (fun() ->
@@ -411,7 +424,6 @@ let test() =
       put R1 (Dir ["x", File "foo"]);
       put R2 (Dir ["x", File "bar"]); sync();
       (* Change contents without changing size and check that change is propagated *)
-      Unix.sleep 2; (* in case time granularity is coarse on this FS *)
       put R1 (Dir ["x", File "f00"]); sync();
 
       check "3a" R1 (Dir ["x", File "f00"]);
@@ -555,6 +567,77 @@ let test() =
     *)
   end;
 
+  if not bothRootsLocal then
+  begin
+    let localR, remoteR, localRaw =
+      match r1 with
+      | Common.Local, _ -> R1, R2, r1
+      | _ -> R2, R1, r2
+    in
+
+    (* Test RPC function "fingerprintSubfile" *)
+    runtest "RPC: transfer append" [] (fun () ->
+      let prefixLen = 1024 * 1024 + 1 in
+      let len = prefixLen + 31 in
+      let contents = String.make len '.' in
+      let fileName = "bigfile" in
+      let prefixPath = Path.fromString fileName in
+      let (workingDir, _) = Fspath.findWorkingDir (snd localRaw) prefixPath in
+      let prefixName = Path.toString (Os.tempPath ~fresh:false workingDir prefixPath) in
+      put remoteR (Dir [(fileName, File contents)]);
+      put localR (Dir [(prefixName, File (String.sub contents 0 prefixLen))]);
+      sync ();
+      check "1" localR (Dir [(fileName, File contents)]);
+    );
+
+    (* Test RPC function "updateProps" *)
+    runtest "RPC: update props" ["times = true"] (fun () ->
+      let state = [("a", File "x")] in
+      put remoteR (Dir state);
+      put localR (Dir []);
+      sync ();
+      (* Having to sleep here is an unfortunate side-effect of the current
+         Windows limitations-inspired time comparison algorithm which is
+         designed to work on FAT filesystems (2-second granularity). *)
+      Unix.sleep 2;
+      put remoteR (Dir state);
+      sync ();
+      check "1" localR (Dir state);
+    );
+
+    (* Test RPC function "replaceArchive" *)
+    runtest "RPC: replaceArchive" [] (fun () ->
+      put localR (Dir [("n", File "to delete")]);
+      put remoteR (Dir []);
+      sync ();
+      put remoteR (Dir []);
+      sync ();
+      check "1" localR (Dir []);
+    );
+
+    (* Test RPC functions "mkdir" and "setDirProp" *)
+    runtest "RPC: mkdir, setDirProp" [] (fun () ->
+      let state = [("subd", Dir [])] in
+      put localR (Dir state);
+      put remoteR (Dir []);
+      sync ();
+      check "1" remoteR (Dir state);
+    );
+
+    (* Test RPC function "setupTargetPaths" *)
+    runtest "RPC: merge" ["merge = Name ma -> echo x> NEW"; "backupcurr = Name ma"] (fun () ->
+      let result = match Sys.os_type with
+        | "Win32" -> ("ma", File "x\r\n")
+        | _ -> ("ma", File "x\n")
+      in
+      put localR (Dir [("ma", File "a")]);
+      put remoteR (Dir [("ma", File "b")]);
+      sync ();
+      check "1" localR (Dir [result]);
+      check "2" remoteR (Dir [result]);
+    );
+  end;
+
   if !failures = 0 then
     Util.msg "Success :-)\n"
   else
