@@ -317,12 +317,6 @@ let createString name ~category ?(cli_only=false) ?(local=false) ?send default ?
     (fun cell -> Uarg.String (fun s -> set cell s))
     Umarshal.string
 
-let createFspath name ~category ?(cli_only=false) ?(local=false) ?send default ?(deprecated=false) doc fulldoc =
-  createPrefInternal name `STRING category cli_only local send default deprecated doc fulldoc
-    (fun v -> [System.fspathToString v])
-    (fun cell -> Uarg.String (fun s -> set cell (System.fspathFromString s)))
-    System.mfspath
-
 let createStringList name ~category ?(cli_only=false) ?(local=false) ?send ?(deprecated=false) doc fulldoc =
   createPrefInternal name `STRING_LIST category cli_only local send [] deprecated doc fulldoc
     (fun v -> v)
@@ -382,10 +376,9 @@ let rec readAFile ?(fail=true) ?(add_ext=true) filename =
   let path = profilePathname ~add_ext:add_ext filename in
   let locname =
     if add_ext then
-      Printf.sprintf "Profile \"%s\" (file \"%s\")"
-        filename (System.fspathToPrintString path)
+      Printf.sprintf "Profile \"%s\" (file \"%s\")" filename path
     else
-      Printf.sprintf "File \"%s\"" (System.fspathToPrintString path)
+      Printf.sprintf "File \"%s\"" path
   in
   let bom = "\xef\xbb\xbf" in (* BOM: UTF-8 byte-order mark *)
   let rec loop chan lineNum lines =
@@ -411,14 +404,13 @@ let rec readAFile ?(fail=true) ?(add_ext=true) filename =
   match chan, fail with
   | None, true when add_ext ->
       raise (Util.Fatal (Printf.sprintf
-        "Profile %s not found (looking for file %s)"
-        filename (System.fspathToPrintString path)))
+        "Profile %s not found (looking for file %s)" filename path))
   | None, true ->
       raise (Util.Fatal (Printf.sprintf
-        "Preference file %s not found"
-        (System.fspathToPrintString path)))
+        "Preference file %s not found" path))
   | None, false -> []
-  | Some chan, _ -> loop chan 1 []
+  | Some chan, _ ->
+      try loop chan 1 [] with e -> close_in_noerr chan; raise e
 
 (* Takes a list of strings in reverse order and yields a list of "parsed lines"
    in correct order *)
@@ -696,6 +688,7 @@ let printFullManDocs () =
   let parRe = Str.regexp "\\\\par *" in
   let underRe = Str.regexp "\\\\_ *" in
   let dollarRe = Str.regexp "\\\\\\$ *" in
+  let dquotRe = Str.regexp "\"" in
   let nn1Re = Str.regexp "\\(\\( -NN-\\)+ -NN-\\|\\( -NN-\\)* -NS-\\)\\." in
   let nn2Re = Str.regexp "\\( -NN-\\)+" in
   let substMacro m s =
@@ -734,6 +727,7 @@ let printFullManDocs () =
     Str.global_replace parRe "\n" >>>
     Str.global_replace underRe "_" >>>
     Str.global_replace dollarRe "$" >>>
+    Str.global_replace dquotRe "\\&\"" >>>
     Str.global_replace nn1Re " Ns " >>>
     Str.global_replace nn2Re "\n" >>>
     Str.global_replace newlineRe "\n" >>>
@@ -779,7 +773,7 @@ let addLine l =
     debug (fun() ->
       Util.msg "Adding '%s' to %s\n" l (System.fspathToDebugString filename));
     let resultmsg =
-      l ^ "' added to profile " ^ System.fspathToPrintString filename in
+      l ^ "' added to profile " ^ filename in
     let ochan =
       System.open_out_gen [Open_wronly; Open_creat; Open_append] 0o600 filename
     in
