@@ -57,19 +57,23 @@ let stringSetFromList l =
 (*                    Debugging / error messages                             *)
 (*****************************************************************************)
 
-let infos = ref ""
+type infos = { s : string; clr : string }
+let infos = ref { s = ""; clr = "" }
 
 let clear_infos () =
-  if !infos <> "" then begin
+  if !infos.clr <> "" then begin
+    print_string !infos.clr;
+    flush stdout
+  end else if !infos.s <> "" then begin
     print_string "\r";
-    print_string (String.make (String.length !infos) ' ');
+    print_string (String.make (String.length !infos.s) ' ');
     print_string "\r";
     flush stdout
   end
 let show_infos () =
-  if !infos <> "" then begin print_string !infos; flush stdout end
-let set_infos s =
-  if s <> !infos then begin clear_infos (); infos := s; show_infos () end
+  if !infos.s <> "" then begin print_string !infos.s; flush stdout end
+let set_infos ?(clr = "") s =
+  if s <> !infos.s then begin clear_infos (); infos := {s; clr}; show_infos () end
 
 let msg f =
   clear_infos ();
@@ -272,6 +276,19 @@ let process_status_to_string = function
   | Unix.WSIGNALED i -> Printf.sprintf "Killed by signal %d" i
   | Unix.WSTOPPED i  -> Printf.sprintf "Stopped by signal %d" i
 
+
+let blockSignals sigs f =
+  let (prevMask, ok) =
+    try (Unix.sigprocmask SIG_BLOCK sigs, true)
+    with Invalid_argument _ -> ([], false) in
+  let restoreMask () =
+    if ok then Unix.sigprocmask SIG_SETMASK prevMask |> ignore in
+  try let r = f () in restoreMask (); r
+  with e ->
+    let origbt = Printexc.get_raw_backtrace () in
+    restoreMask ();
+    Printexc.raise_with_backtrace e origbt
+
 (*****************************************************************************)
 (*                         OS TYPE                                           *)
 (*****************************************************************************)
@@ -328,6 +345,23 @@ let percentageOfTotal current total =
 
 let percent2string p = Printf.sprintf "%3d%%" (truncate (max 0. (min 100. p)))
 
+let gib = 1073741824.
+let mib = 1048576.
+let kib = 1024.
+let bytes2string v =
+  if v > 1_048_051_711L then
+    Printf.sprintf "%.2f GiB" (Int64.to_float v /. gib)
+  else if v > 104_805_171L then
+    Printf.sprintf "%.0f MiB" (Int64.to_float v /. mib)
+  else if v > 1_023_487L then
+    Printf.sprintf "%.1f MiB" (Int64.to_float v /. mib)
+  else if v > 102_348L then
+    Printf.sprintf "%.0f KiB" (Int64.to_float v /. kib)
+  else if v > 999L then
+    Printf.sprintf "%.1f KiB" (Int64.to_float v /. kib)
+  else
+    Printf.sprintf "%Ld B" v
+
 let extractValueFromOption = function
     None -> raise (Fatal "extractValueFromOption failed")
   | Some(v) -> v
@@ -461,7 +495,7 @@ let padto n s = s ^ (String.make (max 0 (n - String.length s)) ' ')
 (*              Building pathnames in the user's home dir                    *)
 (*****************************************************************************)
 
-let homeDirStr = 
+let homeDir () =
     (if (osType = `Unix) || isCygwin then
        safeGetenv "HOME"
      else if osType = `Win32 then
@@ -482,10 +516,7 @@ let homeDirStr =
      else
        assert false (* osType can't be anything else *))
 
-let homeDir () =
-  System.fspathFromString homeDirStr
-
-let fileInHomeDir n = System.fspathConcat (homeDir ()) n
+let fileInHomeDir n = Filename.concat (homeDir ()) n
 
 (*****************************************************************************)
 (*                       .unison dir                                         *)
@@ -497,7 +528,7 @@ let isMacOSX = isMacOSXPred ()
 
 let unisonDir =
   try
-    System.fspathFromString (System.getenv "UNISON")
+    System.getenv "UNISON"
   with Not_found ->
     let genericName =
       fileInHomeDir (Printf.sprintf ".%s" ProjectInfo.myName) in
@@ -506,11 +537,9 @@ let unisonDir =
     else
       genericName
 
-let unisonDirStr = System.fspathToString unisonDir
-
-let fileInUnisonDir str = System.fspathConcat unisonDir str
+let fileInUnisonDir str = Filename.concat unisonDir str
 
 let fileMaybeRelToUnisonDir n =
   if Filename.is_relative n
   then fileInUnisonDir n
-  else System.fspathFromString n
+  else n
